from math import ceil
import torch
from torch import nn
import torch.nn.functional as F

from einops import rearrange, pack, unpack

def exists(val):
    return val is not None

def eval_decorator(fn):
    def inner(self, *args, **kwargs):
        was_training = self.training
        self.eval()
        out = fn(self, *args, **kwargs)
        self.train(was_training)
        return out
    return inner

# nucleus

def top_p(logits, thres = 0.9):
    sorted_logits, sorted_indices = torch.sort(logits, descending=True)
    cum_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

    sorted_indices_to_remove = cum_probs > (1 - thres)
    sorted_indices_to_remove[:, 1:] = sorted_indices_to_remove[:, :-1].clone()
    sorted_indices_to_remove[:, 0] = 0

    sorted_logits[sorted_indices_to_remove] = float('-inf')
    return sorted_logits.scatter(1, sorted_indices, sorted_logits)

# topk

def top_k(logits, thres = 0.9):
    k = ceil((1 - thres) * logits.shape[-1])
    val, ind = torch.topk(logits, k)
    probs = torch.full_like(logits, float('-inf'))
    probs.scatter_(1, ind, val)
    return probs

# top_a

def top_a(logits, min_p_pow=2.0, min_p_ratio=0.02):
    probs = F.softmax(logits, dim=-1)
    limit = torch.pow(torch.max(probs), min_p_pow) * min_p_ratio
    logits[probs < limit] = float('-inf')
    logits[probs >= limit] = 1
    return logits

# autoregressive wrapper class

class AutoregressiveWrapper(nn.Module):
    def __init__(
        self,
        net,
        ignore_index = -100,
        pad_value = 0,
        mask_prob = 0.
    ):
        super().__init__()
        self.pad_value = pad_value
        self.ignore_index = ignore_index

        self.net = net
        self.max_seq_len = net.max_seq_len

        # paper shows masking (MLM) in conjunction with autoregressive decoder-only training leads to big improvements https://arxiv.org/abs/2210.13432
        assert mask_prob < 1.
        self.mask_prob = mask_prob

    @torch.no_grad()
    @eval_decorator
    def generate(
        self,
        start_tokens,
        seq_len,
        eos_token = None,
        temperature = 1.,
        filter_logits_fn = top_k,
        filter_thres = 0.9,
        min_p_pow = 2.0,
        min_p_ratio = 0.02,
        **kwargs
    ):
        device = start_tokens.device
        num_dims = start_tokens.ndim

        start_tokens, ps = pack([start_tokens], '* n')

        b, t = start_tokens.shape

        out = start_tokens

        for _ in range(seq_len):
            x = out[:, -self.max_seq_len:]

            logits = self.net(x, **kwargs)[:, -1]

            if filter_logits_fn in {top_k, top_p}:
                filtered_logits = filter_logits_fn(logits, thres = filter_thres)
                probs = F.softmax(filtered_logits / temperature, dim=-1)

            elif filter_logits_fn is top_a:
                filtered_logits = filter_logits_fn(logits, min_p_pow = min_p_pow, min_p_ratio= min_p_ratio)
                probs = F.softmax(filtered_logits / temperature, dim=-1)

            sample = torch.multinomial(probs, 1)

            out = torch.cat((out, sample), dim=-1)

            if exists(eos_token):
                is_eos_tokens = (out == eos_token)

                if is_eos_tokens.any(dim = -1).all():
                    # mask out everything after the eos tokens
                    shifted_is_eos_tokens = F.pad(is_eos_tokens, (1, -1))
                    mask = shifted_is_eos_tokens.float().cumsum(dim = -1) >= 1
                    out = out.masked_fill(mask, self.pad_value)
                    break

        out = out[:, t:]

        out, = unpack(out, ps, '* n')

        return out

    def forward(self, x, **kwargs):
        seq, ignore_index = x.shape[1], self.ignore_index

        inp, target = x[:, :-1], x[:, 1:]
        inp = torch.where(inp == ignore_index, self.pad_value, inp)

        if self.mask_prob > 0.:
            rand = torch.randn(inp.shape, device = x.device)
            rand[:, 0] = -torch.finfo(rand.dtype).max # first token should not be masked out
            num_mask = min(int(seq * self.mask_prob), seq - 1)
            indices = rand.topk(num_mask, dim = -1).indices
            mask = ~torch.zeros_like(inp).scatter(1, indices, 1.).bool()
            kwargs.update(self_attn_context_mask = mask)

        logits = self.net(inp, **kwargs)

        loss = F.cross_entropy(
            rearrange(logits, 'b n c -> b c n'),
            target,
            ignore_index = ignore_index
        )

        return loss


from pali.transformer import ViTransformerWrapper, Encoder, XTransformer

class VitModel:
    def __init__(self, 
                 image_size=256, 
                 patch_size=32, 
                 dim=512, 
                 depth=6, 
                 heads=8, 
                 *args, **kwargs):
        self.image_size = image_size
        self.patch_size = patch_size
        self.dim = dim

        self.depth = depth
        self.heads = heads
        self.vit = ViTransformerWrapper(
            image_size=image_size,
            patch_size=patch_size,
            attn_layers=Encoder(
                dim=dim,
                depth=depth,
                heads=heads
            )
        )
    
        
    def process(self, img):
        if img is None:
            raise ValueError('Input image cannot be None')
        if img.shape[1:] != (3, self.image_size, self.image_size):
            raise ValueError('Input image must have the shape [*, 3, {}, {}]'.format(self.image_size, self.image_size))
        
        return self.vit(img, return_embeddings=True)
    
class Transformer:
    def __init__(
            self, 
            dim=512, 
            enc_num_tokens=256, 
            enc_depth=6, 
            enc_heads=8, 
            enc_max_seq_len=1024, 
            dec_num_tokens=256, 
            dec_depth=6, 
            dec_heads=8, 
            dec_max_seq_len=1024
        ):
        super().__init__()
        self.pali = XTransformer(
            dim=dim,
            enc_num_tokens=enc_num_tokens,
            enc_depth=enc_depth,
            enc_heads=enc_heads,
            enc_max_seq_len=enc_max_seq_len,
            dec_num_tokens=dec_num_tokens,
            dec_depth=dec_depth,
            dec_heads=dec_heads,
            dec_max_seq_len=dec_max_seq_len
        )

    def process(
            self, 
            prompt, 
            output_text, 
            mask, 
            img_embeds
        ):
        if prompt is None or output_text is None or mask is None or img_embeds is None:
            raise ValueError('None of the input parameters can be None')

        return self.pali(
            prompt,
            output_text,
            mask=mask,
            src_prepend_embeds=img_embeds
        )
    



class Pali:
    def __init__(
        self,
        image_size=256,
        patch_size=32,
        dim=512,
        depth=6,
        heads=8,
        enc_num_tokens=256,
        enc_max_seq_len=1024,
        dec_num_tokens=256,
        dec_max_seq_len=1024
    ):
        self.vit_model = VitModel(
            image_size=image_size,
            patch_size=patch_size,
            dim=dim,
            depth=depth,
            heads=heads
        )

        self.pali_model = Transformer(
            dim=dim,
            enc_num_tokens=enc_num_tokens,
            enc_depth=depth,
            enc_heads=heads,
            enc_max_seq_len=enc_max_seq_len,
            dec_num_tokens=dec_num_tokens,
            dec_depth=depth,
            dec_heads=heads,
            dec_max_seq_len=dec_max_seq_len
        )

    def process(
        self,
        img,
        prompt,
        output,
        mask
    ):
        img_embeds = self.vit_model.process(img)
        result = self.pali_model.process(prompt, output, mask, img_embeds)
        return result





# # usage
# vit_module = ViTModule()
# pali_module = PaliModule()

# #training data
# img = torch.randn(1, 3, 256, 256)
# prompt = torch.randint(0, 256, (1, 1024)) # prompt
# prompt_mask = torch.ones(1, 1024).bool()
# output_text = torch.randint(0, 256, (1, 1024)) #target output text

# img_embeds = vit_module.process(img)

# loss = pali_module.process(prompt, output_text, prompt_mask, img_embeds)

# loss.backward()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pali']

package_data = \
{'': ['*']}

install_requires = \
['primus', 'torch', 'transformers']

setup_kwargs = {
    'name': 'pali-torch',
    'version': '0.0.4',
    'description': 'Pali - PyTorch',
    'long_description': '[![Multi-Modality](agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n\n# Pali: A Multimodal Model\n[![GitHub issues](https://img.shields.io/github/issues/kyegomez/pali)](https://github.com/kyegomez/pali/issues) \n[![GitHub forks](https://img.shields.io/github/forks/kyegomez/pali)](https://github.com/kyegomez/pali/network) \n[![GitHub stars](https://img.shields.io/github/stars/kyegomez/pali)](https://github.com/kyegomez/pali/stargazers) [![GitHub license](https://img.shields.io/github/license/kyegomez/pali)](https://github.com/kyegomez/pali/blob/master/LICENSE)\n[![Share on Twitter](https://img.shields.io/twitter/url/https/twitter.com/cloudposse.svg?style=social&label=Share%20%40kyegomez/pali)](https://twitter.com/intent/tweet?text=Excited%20to%20introduce%20pali,%20the%20all-new%20robotics%20model%20with%20the%20potential%20to%20revolutionize%20automation.%20Join%20us%20on%20this%20journey%20towards%20a%20smarter%20future.%20%23RT1%20%23Robotics&url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fpali)\n[![Share on Facebook](https://img.shields.io/badge/Share-%20facebook-blue)](https://www.facebook.com/sharer/sharer.php?u=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fpali)\n[![Share on LinkedIn](https://img.shields.io/badge/Share-%20linkedin-blue)](https://www.linkedin.com/shareArticle?mini=true&url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fpali&title=Introducing%20pali%2C%20the%20All-New%20Robotics%20Model&summary=pali%20is%20the%20next-generation%20robotics%20model%20that%20promises%20to%20transform%20industries%20with%20its%20intelligence%20and%20efficiency.%20Join%20us%20to%20be%20a%20part%20of%20this%20revolutionary%20journey%20%23RT1%20%23Robotics&source=)\n![Discord](https://img.shields.io/discord/999382051935506503)\n[![Share on Reddit](https://img.shields.io/badge/-Share%20on%20Reddit-orange)](https://www.reddit.com/submit?url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fpali&title=Exciting%20Times%20Ahead%20with%20pali%2C%20the%20All-New%20Robotics%20Model%20%23RT1%20%23Robotics) [![Share on Hacker News](https://img.shields.io/badge/-Share%20on%20Hacker%20News-orange)](https://news.ycombinator.com/submitlink?u=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fpali&t=Exciting%20Times%20Ahead%20with%20pali%2C%20the%20All-New%20Robotics%20Model%20%23RT1%20%23Robotics)\n[![Share on Pinterest](https://img.shields.io/badge/-Share%20on%20Pinterest-red)](https://pinterest.com/pin/create/button/?url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fpali&media=https%3A%2F%2Fexample.com%2Fimage.jpg&description=pali%2C%20the%20Revolutionary%20Robotics%20Model%20that%20will%20Change%20the%20Way%20We%20Work%20%23RT1%20%23Robotics)\n[![Share on WhatsApp](https://img.shields.io/badge/-Share%20on%20WhatsApp-green)](https://api.whatsapp.com/send?text=I%20just%20discovered%20pali,%20the%20all-new%20robotics%20model%20that%20promises%20to%20revolutionize%20automation.%20Join%20me%20on%20this%20exciting%20journey%20towards%20a%20smarter%20future.%20%23RT1%20%23Robotics%0A%0Ahttps%3A%2F%2Fgithub.com%2Fkyegomez%2Fpali)\n\n\nThe open source implementation of the Multi-Modality AI model from ["PaLI: Scaling Language-Image Learning in 100+ Languages"](https://arxiv.org/abs/2209.06794)\n\n## 🌟 Appreciation\nBig bear hugs 🐻💖 to *LucidRains* for the fab x_transformers and for championing the open source AI cause.\n\n## 🚀 Quick Start\n\n```bash\npip install pali-torch\n```\n---\n\n## 🧙 Usage \n```python\nimport torch\nfrom pali.model import Pali\n\nmodel = Pali()\n\nimg = torch.randn(1, 3, 256, 256)\nprompt = torch.randint(0, 256, (1, 1024))\nmask = torch.ones(1, 1024).bool()\noutput_text = torch.randint(0, 256, (1, 1024))\n\nresult = model.process(img, prompt, output_text, mask)\nprint(result)\n\nresult = result.backward()\nprint(result)\n```\n----\n\n# Datasets Strategy\nDataset strategy as closely shown in the paper.\n\nHere is a markdown table with metadata and links to the datasets on HuggingFace for the datasets used:\n\n| Dataset | Description | Size | Languages | Link |\n|-|-|-|-|-|  \n| WebLI | Large-scale web crawled image-text dataset | 10B images, 12B captions | 109 languages | Private |\n| CC3M | Conceptual Captions dataset | 3M image-text pairs | English | [Link](https://huggingface.co/datasets/conceptual_captions) |  \n| CC3M-35L | Translated version of CC3M to 35 languages | 105M image-text pairs | 36 languages | Private |\n| VQAv2 | VQA dataset built on COCO images | 204K images, 1.1M QA pairs | English | [Link](https://huggingface.co/datasets/vqa_v2) |  \n| VQ2A-CC3M | VQA dataset built from CC3M | 3M image-text pairs | English | Private | \n| VQ2A-CC3M-35L | Translated version of VQ2A-CC3M to 35 languages | 105M image-text pairs | 36 languages | Private |\n| Open Images | Large scale image dataset | 9M images with labels | English | [Link](https://huggingface.co/datasets/open_images_v4) |\n| Visual Genome | Image dataset with dense annotations | 108K images with annotations | English | [Link](https://huggingface.co/datasets/visual_genome) |\n| Object365 | Image dataset for object detection | 500K images with labels | English | Private |\n\nThe key datasets used for pre-training PaLI include:\n\n- WebLI: A large-scale multilingual image-text dataset crawled from the web, comprising 10B images and 12B captions in 109 languages.\n\n- CC3M-35L: CC3M Conceptual Captions dataset machine translated into 35 additional languages, totaling 105M image-text pairs in 36 languages. \n\n- VQ2A-CC3M-35L: VQA dataset based on CC3M, also translated into 35 languages. \n\nThe model was evaluated on diverse tasks using standard datasets like VQAv2, Open Images, COCO Captions etc. Links and details provided above.\n\n----\n\n# Todo\n\n- [ ] Make a table of datasets used in paper,\n- [ ] Provide training script\n- [ ] Provide usage/inference scripts\n\n----\n\n## 🎉 Features\n- **Double the Power**: MT5 for text and ViT for images - Pali\'s the superhero we didn\'t know we needed! 💪📖🖼️\n- **Winning Streak**: With roots in the tried-and-true MT5 & ViT, success is in Pali\'s DNA. 🏆\n- **Ready, Set, Go**: No fuss, no muss! Get Pali rolling in no time. ⏱️\n- **Easy-Peasy**: Leave the heavy lifting to Pali and enjoy your smooth sailing. 🛳️\n\n\n## 🌆 Real-World Use-Cases\n\n- **E-commerce**: Jazz up those recs! Understand products inside-out with images & descriptions. 🛍️\n- **Social Media**: Be the smart reply guru for posts with pics & captions. 📱\n- **Healthcare**: Boost diagnostics with insights from images & textual data. 🏥\n\n----\n\n## 📜 License\nMIT\n\n----\n\n## 📚 Citation\n\n```\n@inproceedings{chen2022pali,\n  title={PaLI: Scaling Language-Image Learning in 100+ Languages},\n  author={Chen, Xi and Wang, Xiao},\n  booktitle={Conference on Neural Information Processing Systems (NeurIPS)},\n  year={2022}\n}\n```',
    'author': 'Kye Gomez',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/Pali',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
